#ifndef WINDEF_H
#define WINDEF_H

#ifdef __linux__ 

#include <inttypes.h>
#include <sys/types.h>
#include <linux/limits.h>	// for PATH_MAX
#include <stdlib.h>
#include <strings.h>
#include <wctype.h>
#include <cstring>
#include <cwchar>
#include <cerrno>
#include <pthread.h>

#if defined(__Windows64__) || defined(__linux64__)
#define __stdcall
#else
#define __stdcall  __attribute__((__stdcall__))
#endif

#define __interface		  	struct
#define MAX_PATH          	PATH_MAX
#define INFINITE          	0xFFFFFFFF  // Infinite timeout

#define WINAPI

typedef wchar_t 		 	WCHAR;
typedef wchar_t				__wchar_t;

#define CONST   			const
#define VOID 				void

#if defined(__Windows64__) || defined(__linux64__)
typedef int64_t 			INT_PTR;
typedef uint64_t 			UINT_PTR;
#else
typedef int 				INT_PTR;
typedef unsigned int                    UINT_PTR;
#endif
typedef INT_PTR* 			PINT_PTR;
typedef UINT_PTR* 			PUINT_PTR;

typedef char 				CHAR;
typedef  CHAR *NPSTR, *LPSTR, *PSTR;
typedef  PSTR *PZPSTR;
typedef  CONST PSTR *PCZPSTR;
typedef  CONST CHAR *LPCSTR, *PCSTR;
typedef  PCSTR *PZPCSTR;


typedef WCHAR *PWCHAR, *LPWCH, *PWCH;
typedef CONST WCHAR *LPCWCH, *PCWCH;
typedef WCHAR *NWPSTR, *LPWSTR, *PWSTR;
typedef PWSTR *PZPWSTR;
typedef CONST PWSTR *PCZPWSTR;
typedef CONST WCHAR *LPCWSTR, *PCWSTR;
typedef PCWSTR *PZPCWSTR;

typedef WCHAR _TCHAR;	// TCHAR is defined by MS to be either WCHAR or CHAR
						// depending on the native char type of the system.
						// We define it as WCHAR as the modern windows default
typedef _TCHAR *LPTSTR;
typedef _TCHAR *PTSTR;

#if defined(_UNICODE) || defined(UNICODE)
typedef CONST WCHAR *LPCTSTR;
#else
typedef CONST CHAR *LPCTSTR;
#endif

typedef uint32_t ULONG_PTR;

typedef uint32_t		DWORD;
typedef unsigned char		UCHAR;
typedef bool			BOOL;
typedef unsigned char		BYTE;
typedef unsigned short		WORD;
typedef float			FLOAT;
typedef FLOAT			*PFLOAT;
typedef BOOL			*PBOOL;
typedef BOOL			*LPBOOL;
typedef BYTE			*PBYTE;
typedef BYTE			*LPBYTE;
typedef int			*PINT;
typedef int			*LPINT;
typedef WORD			*PWORD;
typedef WORD			*LPWORD;
typedef uint32_t		*LPLONG;
typedef DWORD			*PDWORD;
typedef DWORD			*LPDWORD;
typedef void			*LPVOID;
typedef CONST void		*LPCVOID;
typedef void			*PVOID;
typedef int32_t			LONG;
typedef DWORD			ULONG;
typedef int			INT;
typedef unsigned int		UINT;
typedef unsigned int		*PUINT;
typedef void			*HMODULE;
typedef void			*HANDLE;
typedef short int		SHORT;
typedef unsigned short int	USHORT;
typedef CHAR *PCHAR;

// 64 bit types
typedef int64_t 		LONGLONG;
typedef int64_t 		LONG64;
typedef uint64_t 		ULONGLONG;
typedef uint64_t 		ULONG64;
typedef ULONGLONG*		PULONG_PTR;
typedef ULONGLONG		DWORD64;

typedef signed char		INT8;
typedef BYTE			UINT8;
typedef signed short		INT16;
typedef unsigned short int  	UINT16;
typedef LONG			INT32;
typedef ULONG			UINT32;
typedef LONGLONG		INT64;
typedef ULONGLONG		UINT64;

#define HRESULT			UINT32

typedef INT8			__int8;
typedef INT16			__int16;
typedef INT32			__int32;
typedef INT64			__int64;

#define FALSE 			0
#define TRUE 			1

#define	NO_ERROR			 0
#define	ERROR_INVALID_OPERATION		-1
#define	ERROR_INVALID_HANDLE		-2
#define	ERROR_INVALID_DATA		-3
#define	ERROR_INVALID_BLOCK		-4
#define	ERROR_NOT_ENOUGH_MEMORY		-5
#define	ERROR_LD_BIND_NOT_DEFINED	-6

#define	_isctype	__isctype
#define _BLANK		_ISblank
#define _SPACE		_ISspace

typedef ULONG_PTR DWORD_PTR, *PDWORD_PTR;

#define MAKELONG(lsw, msw)	((LONG)(((WORD)(((DWORD)(lsw)) & 0xffff)) | ((DWORD)((WORD)(((DWORD)(msw)) & 0xffff))) << 16))
#define LOWORD(l)		((WORD)(((DWORD)(l)) & 0xffff))
#define HIWORD(l)		((WORD)((((DWORD)(l)) >> 16) & 0xffff))

#define MAKEWORD(lsb, msb)	((WORD)(((BYTE)((lsb) & 0xff)) | (((WORD)((BYTE)(msb))) << 8)))
#define LOBYTE(w)		((BYTE)((w) & 0xff))
#define HIBYTE(w)		((BYTE)(((w) >> 8) & 0xff))

/*
** windefs.h is used for both IsolatorCore and Isolator2Acceptance linux adoption
** for the second case, ORG isnt' needed
*/
#ifndef ISOLATOR_EXPORTS

#define sprintf_s(buffer, buffer_size, stringbuffer, ...) snprintf(buffer, buffer_size, stringbuffer, __VA_ARGS__)
#define strcpy_s(dst,size,src) strncpy(dst,src,size)
#define wcscpy_s(dst,size,src) wcsncpy(dst,src,size)
#define _stricmp(str1,str2)    strcasecmp(str1,str2)

#else

#define	sprintf_s(buffer, buffer_size, stringbuffer, ...) (ORG_RET(snprintf(buffer, buffer_size, stringbuffer, __VA_ARGS__)))
#define strcpy_s(dst,size,src) ORG(strncpy(dst,src,size))
#define wcscpy_s(dst,size,src) ORG(wcsncpy(dst,src,size))
#define _stricmp(str1,str2)    ORG_RET(strcasecmp(str1,str2))

#endif

// Dummy for now
typedef struct _EXCEPTION_POINTERS {
    DWORD Exception;
} EXCEPTION_POINTERS;
typedef EXCEPTION_POINTERS *PEXCEPTION_POINTERS;

typedef LONG (*PTOP_LEVEL_EXCEPTION_FILTER)(struct _EXCEPTION_POINTERS *ExceptionInfo);
typedef PTOP_LEVEL_EXCEPTION_FILTER LPTOP_LEVEL_EXCEPTION_FILTER;


#define	_purecall		__cxa_pure_virtual
#define	__declspec(naked)

#define PURE = 0

/* PDB Agent */

typedef struct _SYMBOL_INFO {
    ULONG       SizeOfStruct;
    ULONG       TypeIndex;        // Type Index of symbol
    ULONG64     Reserved[2];
    ULONG       Index;
    ULONG       Size;
    ULONG64     ModBase;          // Base Address of module comtaining this symbol
    ULONG       Flags;
    ULONG64     Value;            // Value of symbol, ValuePresent should be 1
    ULONG64     Address;          // Address of symbol including base address of module
    ULONG       Register;         // register holding value or pointer to value
    ULONG       Scope;            // scope of the symbol
    ULONG       Tag;              // pdb classification
    ULONG       NameLen;          // Actual length of name
    ULONG       MaxNameLen;
    CHAR        Name[1];          // Name of symbol
} SYMBOL_INFO, *PSYMBOL_INFO;

#define CALLBACK
typedef BOOL
(CALLBACK *PSYM_ENUMERATESYMBOLS_CALLBACK)(
    PSYMBOL_INFO pSymInfo,
    ULONG SymbolSize,
    PVOID UserContext
    );

typedef union _ULARGE_INTEGER {
    struct {
        DWORD LowPart;
        DWORD HighPart;
    };
    struct {
        DWORD LowPart;
        DWORD HighPart;
    } u;
#ifdef __linux64__
    struct {
        ULONGLONG QuadPart;
        ULONGLONG HighQuadPart;
    };
#else
    ULONGLONG QuadPart;
#endif
} ULARGE_INTEGER;

typedef union _LARGE_INTEGER {
    struct {
        DWORD LowPart;
        LONG HighPart;
    } DUMMYSTRUCTNAME;
    struct {
        DWORD LowPart;
        LONG HighPart;
    } u;
    LONGLONG QuadPart;
} LARGE_INTEGER;

typedef LARGE_INTEGER* PLARGE_INTEGER;
// parameter nature attributes
#define _In_
#define _In_opt_
#define _Out_
#define _Out_opt_

// function attributes
#define __forceinline __attribute__((always_inline)) inline



// Exceptions stuff
#define EXCEPTION_EXECUTE_HANDLER 1

// DllMAin definitions
#define DLL_PROCESS_ATTACH	1
#define DLL_PROCESS_DETACH	0

// General RTL support function conversion
#define	_countof(a)	(sizeof(a)/sizeof(a[0]))
#define	_mkdir(s)	mkdir(s,0777)
#define	_strdup(s)	strdup(s)
#define	vsprintf_s	vsnprintf

//from imagehelp.h (windows pdb)

#define MAXIMUM_WAIT_OBJECTS    64
#define WAIT_FAILED             ((DWORD)0xFFFFFFFF)
#define TH32CS_SNAPTHREAD       ((DWORD)0x00000004)
#define TH32CS_SNAPPROCESS      ((DWORD)0x00000002) 
#define INVALID_HANDLE_VALUE    ((HANDLE)0xFFFFFFFF)

typedef struct tagTHREADENTRY32 {
  DWORD dwSize;
  DWORD cntUsage;
  pthread_t th32ThreadID;
  pid_t th32OwnerProcessID;
  LONG  tpBasePri;
  LONG  tpDeltaPri;
  DWORD dwFlags;
} THREADENTRY32, *PTHREADENTRY32, *LPTHREADENTRY32;

typedef struct _SECURITY_ATTRIBUTES {
  DWORD  nLength;
  LPVOID lpSecurityDescriptor;
  BOOL   bInheritHandle;
} SECURITY_ATTRIBUTES, *PSECURITY_ATTRIBUTES, *LPSECURITY_ATTRIBUTES;

#define ARRAYSIZE(A) (sizeof(A)/sizeof(A[0]))
#define MIN(A,B)     ((A)<(B)?(A):(B))

inline int _wtoi(const wchar_t *str) {
  return (int)wcstol(str, 0, 10);
}

inline wchar_t *_wcslwr(wchar_t *str) {
  for(wchar_t *wc = str; *wc; ++wc) {
      *wc = towlower(*wc);
  }
  return str;
}

inline int strncpy_s(char *strDest,
                     size_t strDestLen,
                     const char *strSource,
                     size_t count) {

   const size_t len = MIN(strDestLen, count);
   strncpy(strDest, strSource, len);
   strDest[len-1] = 0;
   return ERANGE;
}

#define HKEY HANDLE
#define REGSAM DWORD
#define PHKEY  HKEY*

#define WAIT_TIMEOUT ((DWORD)0x00000102L)

// since linux has no windows-like registry, we use all NULLs here 
#define HKEY_CLASSES_ROOT   NULL
#define HKEY_CURRENT_CONFIG NULL
#define HKEY_CURRENT_USER   NULL
#define HKEY_LOCAL_MACHINE  NULL
#define HKEY_USERS          NULL

#define KEY_READ         0x20019
#define KEY_WRITE        0x20006

#define ERROR_SUCCESS    0u

typedef struct _SYSTEMTIME {
    WORD wYear;
    WORD wMonth;
    WORD wDayOfWeek;
    WORD wDay;
    WORD wHour;
    WORD wMinute;
    WORD wSecond;
    WORD wMilliseconds;
} SYSTEMTIME, *PSYSTEMTIME, *LPSYSTEMTIME;

typedef struct _FILETIME {
  DWORD dwLowDateTime;
  DWORD dwHighDateTime;
} FILETIME, *PFILETIME, *LPFILETIME;

#define HINSTANCE HANDLE
#define HWND      HANDLE

#define REG_QWORD        5u
#define _CRTDBG_CHECK_DEFAULT_DF  0u
#define UTF8_SEQUENCE_MAXLEN 6

#if defined(__Windows64__) || defined(__linux64__)
  typedef uint64_t SIZE_T;
#else
  typedef unsigned long SIZE_T;
#endif

typedef SIZE_T *PSIZE_T;

typedef struct _MEMORY_BASIC_INFORMATION {
  PVOID  BaseAddress;
  PVOID  AllocationBase;
  DWORD  AllocationProtect;
  SIZE_T RegionSize;
  DWORD  State;
  DWORD  Protect;
  DWORD  Type;
} MEMORY_BASIC_INFORMATION, *PMEMORY_BASIC_INFORMATION;

#define PAGE_READWRITE (PROT_READ|PROT_WRITE)

#define WAIT_OBJECT_0 0

typedef pthread_mutex_t CRITICAL_SECTION;
typedef CRITICAL_SECTION* LPCRITICAL_SECTION;

typedef int LPOVERLAPPED;

#define _MAX_PATH   260 // max. length of full pathname

#endif // #ifdef __linux__
#endif // #ifndef WINDEF_H
