#ifdef __linux__

#include "stdafx.h"

#else

#include <string>
#include <exception>
using namespace std;
#include <Isolator.h>
#include <gtest/gtest.h>

#endif

#include "../ClassesUnderTest/Person.h"
class ChainsBehavior : public ::testing::Test
{
public:
	
	virtual void TearDown() 
	{
		ISOLATOR_CLEANUP();
	}
};


	/*****************************************
	Mocking a chain of methods
	******************************************/

	/*
	* This shows the basic rule for mocking a chain of methods of a faked object.
	* the behavior will be as follow:
	*  - Every method in the chain will be mocked
	*  - When the chain of method will be called the behavior will be as modified.
	*/
	TEST_F(ChainsBehavior ,SettingChainedMethodsBehavior)
	{
		Person* fake = FAKE<Person>(FakeOptions::CallOriginal); // Can also use FakeOptions::Lean, but methods will not be tracked and asserting them will require using WHEN_CALLED on the method
		
		WHEN_CALLED(fake->GetAddressLocation(0)->Latitude()).Return(10);

		ASSERT_EQ(10, fake->GetLocationLatitude(0));
	}

	
	/*
	* The same rules as above apply when all the instances of the same object 
	  is  being faked by the FAKE_ALL macro.
	*/

	TEST_F(ChainsBehavior, SettingChainedMethodsBehaviorFakeAll)
	{
		Person* fake = FAKE_ALL<Person>(FakeOptions::CallOriginal); // Can also use FakeOptions::Lean, but methods will not be tracked and asserting them will require using WHEN_CALLED on the method
		WHEN_CALLED(fake->GetAddressLocation(0)->Latitude()).Return(10);

       Person * instance = new Person();

       ASSERT_EQ(10,fake->GetLocationLatitude(0));
	}

	/*
	* The same rules as above apply when the methods in the chain belong to live object. 
	*/

	TEST_F(ChainsBehavior, SettingChainedMethodsBehaviorLive)
	{
		Person* liveObj = new Person();
		
		WHEN_CALLED(liveObj->GetAddressLocation(0)->Latitude()).Return(10);
		
		ASSERT_EQ(10,liveObj->GetLocationLatitude(0));
	}


