#ifdef __linux__

#include "stdafx.h"

#else

#include <string>
#include <exception>
#include <Isolator.h>
#include <gtest/gtest.h>
using namespace std;

#endif

#include "../ClassesUnderTest/Person.h"


class BehaviorSequencing : public ::testing::Test
{
public:
	
	virtual void TearDown() 
	{
		ISOLATOR_CLEANUP();
	}
};

/*****************************************
		Setting behaviors in sequence
******************************************/

/*
 * This shows the basic rule for sequenced behavior.
 * When set the second time the behavior of a method
 * the behavior will be as follow:
 *  - Every call will return the next value that was specified in the order of WHEN_CALLED statements
 *  - The last statement of WHEN_CALLED will define the default for each subsequent call .  
 */
TEST_F(BehaviorSequencing, LastSequencedValueRemains)
{
	Person person;
	Address* fakeAddress = FAKE<Address>();

	WHEN_CALLED(fakeAddress->GetZipCode()).Return (90210);
	WHEN_CALLED(fakeAddress->GetZipCode()).Return (93108);
	WHEN_CALLED(fakeAddress->GetZipCode()).Return (91008);
	

	ASSERT_EQ(90210, person.GetAddressZipCode(fakeAddress));
	ASSERT_EQ(93108, person.GetAddressZipCode(fakeAddress));
	ASSERT_EQ(91008, person.GetAddressZipCode(fakeAddress));
	ASSERT_EQ(91008, person.GetAddressZipCode(fakeAddress));
}

/*
 * The same sequencing rules as above apply to faking different 
 * behaviors on a method 
 */
TEST_F(BehaviorSequencing, MixingBehaviors)
{
	Person person;
	Address* fakeAddress = FAKE<Address>();

	WHEN_CALLED(fakeAddress->GetZipCode()).CallOriginal();
	WHEN_CALLED(fakeAddress->GetZipCode()).Return (90210);
	WHEN_CALLED(fakeAddress->GetZipCode()).ReturnFake ();
	
	
	ASSERT_EQ(-1, person.GetAddressZipCode(fakeAddress));
	ASSERT_EQ(90210, person.GetAddressZipCode(fakeAddress));
	ASSERT_EQ(0, person.GetAddressZipCode(fakeAddress));
	ASSERT_EQ(0, person.GetAddressZipCode(fakeAddress));

}
