#ifdef __linux__

#include "stdafx.h"

#else

#include <string>
#include <exception>
#include <Isolator.h>
#include <gtest/gtest.h>
using namespace std;

#endif

#include "../ClassesUnderTest/Person.h"

class ChangeReturnValue : public ::testing::Test
{
public:
	virtual void TearDown() 
	{
		ISOLATOR_CLEANUP();		
	}
};

/*****************************************
		Returning fake values
******************************************/

/*
 * GetZipCode will return the supplied value instead of running the original implementation
 */
TEST_F(ChangeReturnValue, ReturnPrimitive)
{
	Person person;
	Address* fakeAddress = FAKE<Address>();

	WHEN_CALLED(fakeAddress->GetZipCode()).Return(100);

	ASSERT_EQ(100, person.GetAddressZipCode(fakeAddress));
}

/*
 * GetCity will return the supplied string instead of running the original implementation
 * ReturnString works on single byte strings. Use ReturnStringW for wide-char strings
 */
TEST_F(ChangeReturnValue, ReturnString)
{
	Person person;
	Address* fakeAddress = FAKE<Address>();

	WHEN_CALLED(fakeAddress->GetCity()).Return("New York");

	ASSERT_STREQ("New York", person.GetAddressCity(fakeAddress));
}

/*
 * GetLocation will return the supplied object instead of running the original implementation
 * ReturnPtr takes the address of the returned object
 * The default return of the faked GetLocation is a fake GPSLocation object
 */
TEST_F(ChangeReturnValue, ReturnObject)
{
	Person person;
	Address* fakeAddress = FAKE<Address>();
	
	GPSLocation newyork(41, 74);

	WHEN_CALLED(fakeAddress->GetLocation()).Return(&newyork);
	GPSLocation* result = person.GetAddressLocation(fakeAddress);
	ASSERT_EQ(74, result->Longitude());
	ASSERT_EQ(41, result->Latitude());
}



