#ifdef __linux__

#include "stdafx.h"

#else

#include <string>
#include <exception>
#include <Isolator.h>
#include <gtest/gtest.h>
using namespace std;

#endif

#include "Zombieland.h"
#include "../ClassesUnderTest/Person.h"


// DoInstead() calls the provided function instead of the original implementation
class DoInstead : public ::testing::Test
{
public:
	virtual void TearDown()
	{
		ISOLATOR_CLEANUP();
	}

};

/*******************************************
	Setting custom behavior with DoInstead
********************************************/
	
/*  Sometimes, we need to supply different behavior based on the paramaters that 
	are passed to the method at runtime. In this example, we're returning different 
	values based on different method calls, when the faked method is static.
*/
TEST_F(DoInstead, ReplacingStaticMethodCallBasedOnParameters)
{
	Person person;
	WHEN_CALLED(Address::GetCapitalCityByCountry(ANY_VAL(string))).
		DoStaticOrGlobalInstead(Zombieland::GetZombielandCapitalCity, NULL);

	string zombieCapital = person.GetCapitalCity("Zombieland");
	string otherCapital = person.GetCapitalCity("US");

	ASSERT_EQ(0, zombieCapital.compare("GhostTown"));
	ASSERT_EQ(0, otherCapital.compare("Elsewhere"));
}

/*  In this example, we're returning different 
	values based on the order the methods get called, when the faked method is an instance .
*/
TEST_F(DoInstead, ReplacingInstanceMethodCallBasedOnOrder)
{
	Person person;
	Address* address = FAKE<Address>();
	Zombieland land;

	WHEN_CALLED(address->GetCityByCountry(ANY_VAL(string))).
		DoMemberFunctionInstead(&land, ReturnDifferentCityOnOddAndEvenCalls);

	string firstCity  = person.GetCityFromAddress(address, "");
	string secondCity = person.GetCityFromAddress(address, "");
	string thirdCity  = person.GetCityFromAddress(address, "");
	string fourthCity = person.GetCityFromAddress(address, "");

	ASSERT_EQ(0, firstCity.compare("OddCity"));
	ASSERT_EQ(0, secondCity.compare("EvenCity"));
	ASSERT_EQ(0, thirdCity.compare("OddCity"));
	ASSERT_EQ(0, fourthCity.compare("EvenCity"));
}

#ifdef __linux__
DWORD __stdcall GetTickCount()
{
        return 0;
}
#endif

DWORD __stdcall GetOtherTickCount()
{
	return 10000;
}

/* The following example shows how to replace behavior of global functions. This is helpful in case of Windows API 
   or any C global (free) functions. In this case, we're replacing the behavior of the Windows API GetTickCount. 
*/
TEST_F(DoInstead, ChangeBehaviorToGlobalFunctionWithContext)
{
	FAKE_GLOBAL(GetTickCount);

	WHEN_CALLED(GetTickCount()).DoStaticOrGlobalInstead(GetOtherTickCount, NULL);

	ASSERT_EQ(10000, GetTickCount());
	
}


