#ifdef __linux__

#include "stdafx.h"

#else

#include <string>
#include <exception>
#include <Isolator.h>
#include <gtest/gtest.h>
using namespace std;

#endif

#include "../ClassesUnderTest/Person.h"

// Static calls can be faked explicitely on each method, or on all methods with the FAKE_STATICS macro. 
// All faked static methods follow Recursive Fakes and sequncing rules, like faked instance methods
class FakeStaticMethods : public ::testing::Test
{
public:
	virtual void TearDown()
	{
		ISOLATOR_CLEANUP();
	}
};


/*****************************************
		Faking static methods
******************************************/


/*	It is useful to fake a factory method, and return a controlled object.
	In this example, we create a real object to return, and modify its behavior.
	We then use WHEN_CALLED on the static factory method to return it.
*/
TEST_F(FakeStaticMethods, ReturningAFakeObjectFromAFactoryMethod)
{
	// Preparing the returned object
	Address myAddress;
	WHEN_CALLED(myAddress.GetCity()).Return("London");

	// Faking the static factory method
	WHEN_CALLED(Address::GetSingleton()).Return(&myAddress);
	WHEN_CALLED(Address::ClearSingleton()).Ignore();
	
	// Checking the resulting call
	ASSERT_STREQ("London", Person::GetCityFromAddressFactory());
}

/*	Sometimes we need to fake all the static methods on a class.
	We use the FAKE_STATICS macro for that. All faked methods default
	behaviors follow the Recursive Fakes and sequencing rules.
	In this case, the default char* returned from GetCity is an empty string.
*/
TEST_F(FakeStaticMethods, FakingAllStaticMethodsOnAType)
{
	// Fake all methods
	FAKE_STATICS<Address>();
	
	// All static Address methods are faked.
	ASSERT_STREQ("", Person::GetCityFromAddressFactory());

	// In fact, because the methods were faked, the GetSingleton always returns a fake
	// recursive pointer
	ASSERT_TRUE(Address::GetSingleton() != NULL);
}

/*	In this example, we create clear a singleton and make sure that calling it calls
    a private static method CountReference()
*/
TEST_F(FakeStaticMethods, IgnoringPrivateStaticMethod)
{
	Address* address = new Address();
	PRIVATE_WHEN_CALLED(address,Address::CountReference).Ignore();
	
	// act
	Address::ClearSingleton();

	// assert
	PRIVATE_ASSERT_WAS_CALLED(address,Address::CountReference);
}
