#ifdef __linux__

#include "stdafx.h"

#else

#include <string>
#include <exception>
#include <Isolator.h>
#include <gtest/gtest.h>
using namespace std;

#endif

#include "../ClassesUnderTest/Person.h"


/*	Future objects are object that are created, either on the stack or the heap, after the method under test
	was called. Isolator++ catches the constructor call and fakes it for one or any objects of the type specified
	by the FAKE_ALL macro. The macro also returns a proxy that allows to change behavior and assert method calls 
	on the future object.
*/
class FutureObjectTests : public ::testing::Test
{
public:
	virtual void TearDown()
	{
		ISOLATOR_CLEANUP();
	}
};


/*****************************************
		Faking future objects
******************************************/

/*	FAKE_ALL is effective with future objects, for examples singletons. 
	These are usually created inside factory methods, like in this test. 
*/
TEST_F(FutureObjectTests, FakingTheSingletonObjectAndChangingBehavior)
{
	Address* fakeAddress = FAKE_ALL<Address>();

	WHEN_CALLED(fakeAddress->GetCity()).Return("London");
	WHEN_CALLED(Address::ClearSingleton()).Ignore();

	ASSERT_STREQ("London", Person::GetCityFromAddressFactory());
}

/*	Faking future objects using FAKE_ALL also helps in asserting calls on them.
	We can use any of Isolator's ASSERT APIs on the objects, and see if methods were called.
*/
TEST_F(FutureObjectTests, FakingTheSingletonObjectAndAssertingMethodCalls)
{
	Address* fakeAddress = FAKE_ALL<Address>();

	WHEN_CALLED(Address::ClearSingleton()).Ignore();
	Person::GetCityFromAddressFactory();

	ASSERT_WAS_CALLED(fakeAddress->GetCity());
}

