#ifdef __linux__

#include "stdafx.h"

#else

#include <string>
#include <exception>
#include <Isolator.h>
#include <gtest/gtest.h>
using namespace std;

#endif

#include "../ClassesUnderTest/Person.h"

class GlobalCMethods : public ::testing::Test
{
public:
	virtual void TearDown()
	{
		ISOLATOR_CLEANUP();
	}
};


/*****************************************
		Faking global methods
******************************************/

/*	Faking and asserting on global methods requires using WHEN_CALLED to set its behavior. 
	In cases of multiple definitions it's required first declaring them with FAKE_GLOBAL.
	Using FAKE_GLOBAL will fake all implementations of the method. The default behavior of a method
	declared with FAKE_GLOBAL is returning recursive fakes. In this example
	We require fopen to return NULL, so we use WHEN_CALLED to do that. Since we don't care about 
	what fclose returns, we don't define a behavior.
*/
TEST_F(GlobalCMethods, FakingReturnValueOnGlobalMethod)
{
	FAKE_GLOBAL(fopen);
	FAKE_GLOBAL(fclose);
	WHEN_CALLED(fopen(_, _)).Return(0);

	Person person;
	person.SaveData();

	ASSERT_NOT_CALLED(fclose(_));
}

/*	For the other case, we simulate all global calls: fopen, fwrite and fclose.
	This time, we want fopen to return a non-NULL value. Since the FAKE_GLOBAL macro defines
	a recursive fake, it goes into the if-clause.
*/
TEST_F(GlobalCMethods, FakingMethodAndUsingRecursiveFake)
{
	Person person;

	FAKE_GLOBAL(fopen);
	FAKE_GLOBAL(fwrite);
	FAKE_GLOBAL(fclose);

	person.SaveData();
	
	ASSERT_WAS_CALLED(fclose(_));

}
