#ifdef __linux__

#include "stdafx.h"

#else

#include <string>
#include <exception>
#include <Isolator.h>
#include <gtest/gtest.h>
using namespace std;

#endif

#include "../ClassesUnderTest/Person.h"

// Live objects are partially real, partially faked. We create 
// them as regular objects, and then use WhenCalled to change part of their behavior.
class LiveObjectTests: public ::testing::Test
{
public:
	virtual void TearDown() 
	{
		ISOLATOR_CLEANUP();
	}
};

/*****************************************
		Working with Live Objects
******************************************/

/*	We usually work with live objects when they are under test. In this case, we fake an internal call 
	to a public method. To change that behavior we use the WHEN_CALLED macro on the object we just created.
*/
TEST_F(LiveObjectTests, FakingAPublicMethod)
{
	Person person;
	WHEN_CALLED(person.GetName()).Return("Tommy");

	ASSERT_STREQ("Lassie", person.GetDogName());

}

/*	Asserting a method call uses the ASSERT_WAS_CALLED APIs or similar. However, in order for that to work, we
	need to use at least once an Isolator API on that object before. This can be WHEN_CALLED or anything similar,
	and doesn't have to be for the same method. 
	In this example, we both change the behavior of the public method, and assert it was called.
*/
TEST_F(LiveObjectTests, AssertingACall)
{
	Person person;
	WHEN_CALLED(person.GetName()).Return("Jerry");

	ASSERT_STREQ("Rex", person.GetDogName());
	ASSERT_WAS_CALLED(person.GetName());
}


/*	Sometimes, we need to change behavior on a non-public method of the tested object.
	We use the PRIVATE_WHEN_CALLED macro for that.
*/
TEST_F(LiveObjectTests, FakingAPrivateMethod)
{
	Person* person = new Person();
	PRIVATE_WHEN_CALLED(person, IsValidName).Return(false);
	ASSERT_THROW(person->Save(""), exception);
	delete person;
}

