#ifdef __linux__

#include "stdafx.h"

#else

#include <string>
#include <exception>
#include <Isolator.h>
#include <gtest/gtest.h>
using namespace std;

#endif

#include "../ClassesUnderTest/Person.h"
#include "../ClassesUnderTest/ITransport.h"

//	Faking pure virtual objects is just as easy as faking any other type.
//    use the FAKE macro, this will create an object that implements all the pure virtual methods.
class PureVirtualTests: public ::testing::Test
{
public:
	virtual void TearDown() 
	{
		ISOLATOR_CLEANUP();
	}
};

/*****************************************
		Faking pure virtual methods
******************************************/

/*	FAKE is effective with pure virtual method too
*/
TEST_F(PureVirtualTests, FakingAPureVirutualMethodAndChangingBehavior)
{
	ITransport* fakeTransport = FAKE<ITransport>();

    string bus("Bus");
	WHEN_CALLED(fakeTransport->GetTransport()).Return(&bus);

	Person person;
	person.SetTransport(fakeTransport);

	ASSERT_EQ(0, person.GetTransportName().compare(bus));
}

/*	Faking pure virtual methods using FAKE also helps in asserting calls on them.
	We can use any of Isolator's ASSERT APIs on the objects, and see if methods were called.
*/
TEST_F(PureVirtualTests, FakingAPureVirutualMethodAndAssertingMethodCalls)
{
	ITransport* fakeTransport = FAKE<ITransport>();

    string bus("Bus"); 
	WHEN_CALLED(fakeTransport->GetTransport()).Return(&bus);

	Person person;
	person.SetTransport(fakeTransport);

	auto result = person.GetTransportName();

	ASSERT_WAS_CALLED(fakeTransport->GetTransport());
}



