#ifdef __linux__

#include "stdafx.h"

#else

#include <string>
#include <exception>
#include <Isolator.h>
#include <gtest/gtest.h>
using namespace std;

#endif

#include "../ClassesUnderTest/Person.h"

// ISOLATOR_SET_VARIABLE is used to change variables, usually on the tested object
// ISOLATOR_GET_VARIABLE is used query variables, usually on the tested object
class SettingMembersTests : public ::testing::Test
{
public:
	virtual void TearDown()
	{
		ISOLATOR_CLEANUP();
	}
};


/**************************************************************************************
  Setting static and instance members with ISOLATOR_GET_VARIABLE and ISOLATOR_SET_VARIABLE
***************************************************************************************/

/*	In this test, we set the private m_check member and check 
    The value was actually set. 
*/
TEST_F(SettingMembersTests, SetPrivateInstanceMember)
{
	Person* person = new Person();
	ISOLATOR_SET_VARIABLE(person, m_id, 10);

	// now let's get that private member
	int memberValue;
	ISOLATOR_GET_VARIABLE(person, m_id, memberValue);
	
	ASSERT_EQ(10, memberValue);
	delete person;
}

/*	In this test, we set the private static m_static member and check 
    The value was actually set. 
*/
TEST_F(SettingMembersTests, SetPrivateStaticMember)
{
	ISOLATOR_SET_VARIABLE(_, Person::m_static, 1);

	// now let's get that private member
	int memberValue;
	ISOLATOR_GET_VARIABLE(_, Person::m_static, memberValue);
	
	ASSERT_EQ(1, memberValue);
}


/*	In this test, we'll set and then get the private m_array member and check 
    The value was actually set. 
*/
TEST_F(SettingMembersTests, GetPrivateArray)
{
	Person* person = new Person();
	int arr_set[5] = {1, 2, 3, 4, 5};

	ISOLATOR_SET_VARIABLE(person, m_array, &arr_set[0]);

	// now let's get that private member. 
	//Note, getting an array value will NOT make any copy but return a pointer to the array:
	int* arr_get = NULL;
	ISOLATOR_GET_VARIABLE(person, m_array, arr_get);
	
	for (int i = 0; i < 5; i++)
		ASSERT_EQ(arr_set[i], arr_get[i]);
}
